"""
This module contains methods to construct learning curves, which evaluate some cross-validation performance metric (e.g. RMSE)
as a function of amount of training data (i.e. a sample learning curve) or as a function of the number of features used
in the fitting (i.e. a feature learning curve).
"""

import numpy as np
import pandas as pd
import warnings
import logging
import os

from sklearn.model_selection import learning_curve
from sklearn.feature_selection import f_regression

from mastml.legos import feature_selectors as fs

# Ignore the harmless warning about the gelsd driver on mac.
warnings.filterwarnings(action="ignore", module="scipy",
                        message="^internal gelsd")

log = logging.getLogger('mastml')

def sample_learning_curve(X, y, estimator, cv, scoring, Xgroups=None):
    """
    Method that calculates data used to plot a sample learning curve, e.g. the RMSE of a cross-validation routine using a
    specified model and a given fraction of the total training data

    Args:
        X: (numpy array), array of X data values

        y: (numpy array), array of y data values

        estimator: (scikit-learn model object), a scikit-learn model used for fitting

        cv: (scikit-learn cross validation object), a scikit-learn cross validation object to construct train/test splits

        scoring: (scikit-learn metric object), a scikit-learn metric to use as a scorer

        Xgroups: (list), list of row indices corresponding to each group

    Returns:
        train_sizes: (numpy array), array of fractions of training data used in learning curve

        train_mean: (numpy array), array of means of training data scores for each training data fraction

        test_mean: (numpy array), array of means of testing data scores for each training data fraction

        train_stdev: (numpy array), array of standard deviations of training data scores for each training data fraction

        test_stdev: (numpy array), array of standard deviations of testing data scores for each training data fraction

    """

    train_sizes = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1])
    if Xgroups.shape[0] > 0:
        Xgroups = np.array(Xgroups).reshape(-1, )
    else:
        Xgroups = np.zeros(len(y))

    train_sizes, train_scores, valid_scores = learning_curve(estimator=estimator, X=X, y=y, train_sizes=train_sizes,
                                                             scoring=scoring, cv=cv, groups=Xgroups)
    train_mean = np.mean(train_scores, axis=1)
    test_mean = np.mean(valid_scores, axis=1)
    train_stdev = np.std(train_scores, axis=1)
    test_stdev = np.std(valid_scores, axis=1)

    return train_sizes, train_mean, test_mean, train_stdev, test_stdev

def feature_learning_curve(X, y, estimator, cv, scoring, selector_name, savepath, n_features_to_select=None, Xgroups=None):
    """
    Method that calculates data used to plot a feature learning curve, e.g. the RMSE of a cross-validation routine using a
    specified model and a given number of features

    Args:
        X: (numpy array), array of X data values

        y: (numpy array), array of y data values

        estimator: (scikit-learn model object), a scikit-learn model used for fitting

        cv: (scikit-learn cross validation object), a scikit-learn cross validation object to construct train/test splits

        scoring: (scikit-learn metric object), a scikit-learn metric to use as a scorer

        selector_name: (str), name of a scikit-learn or MAST-ML feature selection routine

        n_features_to_select: (int), total number of features to select, i.e. stopping criterion for number of features

        Xgroups: (list), list of row indices corresponding to each group

    Returns:
        train_sizes: (numpy array), array of fractions of training data used in learning curve

        train_mean: (numpy array), array of means of training data scores for each number of features

        test_mean: (numpy array), array of means of testing data scores for each number of features

        train_stdev: (numpy array), array of standard deviations of training data scores for each number of features

        test_stdev: (numpy array), array of standard deviations of testing data scores for each number of features

    """
    if Xgroups.shape[0] > 0:
        Xgroups = np.array(Xgroups).reshape(-1, )
    else:
        Xgroups = np.zeros(len(y))
    train_mean = list()
    train_stdev = list()
    test_mean = list()
    test_stdev = list()
    if not n_features_to_select:
        n_features_to_select = X.shape[1].tolist()
    train_sizes = range(n_features_to_select)
    train_sizes = [1+f for f in train_sizes]
    features_selected = list()
    n_features = list()
    for feature in train_sizes:
        n_features.append(feature)
        if selector_name == 'RFE':
            log.warning("Using RFE as feature selector does not support a custom CV or grouping scheme. Your learning"
                        "curve will be generated properly, but will not use the custom CV or grouping scheme")
            try:
                Xnew = fs.name_to_constructor[selector_name](estimator, feature).fit(X, y).transform(X)
            except RuntimeError:
                log.error("You have specified an estimator for RFE that does not have a coef_ or feature_importances_ attribute. "
                          "Acceptable models to use with RFE include: LinearRegression, Lasso, SVR, DecisionTreeRegressor, "
                          "RandomForestRegressor, ExtraTreesRegressor, AdaBoostRegressor, etc.")
        elif selector_name == 'SelectKBest':
            log.warning("Using SelectKBest as feature selector does not support a custom estimator model, CV or grouping scheme. "
                        "Your learning curve will be generated properly, but will not use the custom model, CV or grouping scheme")
            Xnew = fs.name_to_constructor[selector_name](f_regression, feature).fit(X, y).transform(X)
        elif selector_name == 'MASTMLFeatureSelector':
            Xnew = fs.name_to_constructor[selector_name](estimator, feature, cv).fit(X, y, pd.DataFrame(Xgroups)).transform(X)
        elif selector_name == 'SequentialFeatureSelector':
            log.warning("Using SequentialFeatureSelector as feature selector does not support a custom CV or grouping scheme. "
                        "Your learning curve will be generated properly, but will not use the custom CV or grouping scheme")
            Xnew = fs.name_to_constructor[selector_name](estimator, feature).fit(pd.DataFrame(X), pd.DataFrame(y)).transform(X)
        elif selector_name == None:
            log.warning("A selector name for learning curve calculation was not found. Defaulting to using the "
                        "MASTMLFeatureSelector for learning curve")
            Xnew = fs.name_to_constructor["MASTMLFeatureSelector"](estimator, feature, cv).fit(X, y, pd.DataFrame(Xgroups)).transform(X)
        else:
            log.error("You have specified an invalid selector_name for learning curve. Either leave blank to use the default"
                      " MASTMLFeatureSelector or use one of SelectKBest, RFE, SequentialFeatureSelector, MASTMLFeatureSelector")
            exit()
        # Need to use arrays to avoid indexing issues when leaving out validation data
        features_selected.append(Xnew.columns.tolist())

        Xnew = np.array(Xnew)
        y = np.array(y)
        Xgroups = np.array(Xgroups)
        cv_number=1
        train_scores = dict()
        test_scores = dict()
        for trains, tests in cv.split(Xnew, y, Xgroups):
            model = estimator.fit(Xnew[trains], y[trains])
            train_vals = model.predict(Xnew[trains])
            test_vals = model.predict(Xnew[tests])
            train_scores[cv_number] = scoring._score_func(train_vals, y[trains])
            test_scores[cv_number] = scoring._score_func(test_vals, y[tests])
            cv_number += 1
        train_mean.append(np.mean(list(train_scores.values())))
        train_stdev.append(np.std(list(train_scores.values())))
        test_mean.append(np.mean(list(test_scores.values())))
        test_stdev.append(np.std(list(test_scores.values())))

    datadict = {"n_features": n_features, "features selected": features_selected}
    pd.DataFrame().from_dict(data=datadict).to_csv(os.path.join(savepath, 'features_selected_in_learning_curve.csv'))

    return np.array(train_sizes), np.array(train_mean), np.array(test_mean), np.array(train_stdev), np.array(test_stdev)



"""
This module provides a name_to_constructor dict for all models/estimators in scikit-learn, plus a couple test models and
error handling functions
"""

import warnings
import inspect

import sklearn.base
import sklearn.utils.testing
import joblib
import numpy as np
import os

# Sometimes xgboost is hard to install so make it optional
try:
    import xgboost as xgb
except:
    pass

import keras
from keras.models import model_from_json
from keras.models import load_model
from keras.models import Sequential

import random

import pandas as pd

#from . import keras_models
from mastml import utils

from scipy import stats

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    name_to_constructor = dict(sklearn.utils.testing.all_estimators())

class AlwaysFive(sklearn.base.RegressorMixin):
    """
    Class used as a test model that always predicts a value of 5.

    Args:

        constant: (int), the value to predict. Always 5 by default

    Methods:

        fit: Just passes through to maintain scikit-learn structure

        predict: Provides predicted model values based on X features

            Args:

                X: (numpy array), array of X features

            Returns:

                (numpy array), prediction array where all values are equal to constant

    """
    def __init__(self, constant = 5):
        self.five = constant

    def fit(self, X, y, groups=None):
        return self

    def predict(self, X):
        return np.array([self.five for _ in range(len(X))])

class RandomGuesser(sklearn.base.RegressorMixin):
    """
    Class used as a test model that always predicts random values for y data.

    Args:

        None

    Methods:

        fit: Constructs possible predicted values based on y data

            Args:

                y: (numpy array), array of y data

        predict: Provides predicted model values based on X features

            Args:

                X: (numpy array), array of X features

            Returns:

                (numpy array), prediction array where all values are random selections of y data

    """
    def __init__(self):
        pass

    def fit(self, X, y, groups=None):
        self.possible_answers = y
        return self

    def predict(self, X):
        return np.random.choice(self.possible_answers, size=X.shape[0])

class KerasRegressor():
    def __init__(self, conf_dict):
        self.conf_dict = conf_dict
        self.model = self.build_model()

    def build_model(self):
        model_vals = self.conf_dict
        model = Sequential()

        for layer_dict, layer_val in model_vals.items():
            if (layer_dict != 'FitParams'):
                layer_type = layer_val.get('layer_type')
                layer_name_asstr = layer_type
                if layer_name_asstr == 'Dense':
                    neuron_num = int(layer_val.get('neuron_num'))
                    if (layer_dict == 'Layer1'):
                        input_dim = int(layer_val.get('input_dim'))
                    kernel_initializer = layer_val.get('kernel_initializer')
                    activation = layer_val.get('activation')
                elif layer_name_asstr == 'Dropout':
                    rate = float(layer_val.get('rate'))
                for layer_name, cls in inspect.getmembers(keras.layers, inspect.isclass):
                    layer_type = getattr(keras.layers, layer_name_asstr)  # (neuron_num)

            else:
                if layer_val.get('rate'):
                    self.rate = float(layer_val.get('rate'))
                if layer_val.get('epochs'):
                    self.epochs = int(layer_val.get('epochs'))
                else:
                    self.epochs = 1
                if layer_val.get('batch_size'):
                    self.batch_size = int(layer_val.get('batch_size'))
                else:
                    self.batch_size = None
                if layer_val.get('loss'):
                    self.loss = str(layer_val.get('loss'))
                else:
                    self.loss = 'mean_squared_error'
                if layer_val.get('optimizer'):
                    self.optimizer = str(layer_val.get('optimizer'))
                else:
                    self.optimizer = 'adam'
                if layer_val.get('metrics'):
                    self.metrics = layer_val.get('metrics').split(',')
                else:
                    self.metrics = ['mae']
                if layer_val.get('verbose'):
                    self.verbose = str(layer_val.get('verbose'))
                else:
                    self.verbose = 0
                if layer_val.get('shuffle'):
                    self.shuffle = bool(layer_val.get('shuffle'))
                else:
                    self.shuffle = True
                if layer_val.get('validation_split'):
                    self.validation_split = float(layer_val.get('validation_split'))
                else:
                    self.validation_split = 0.0
                continue

            if (layer_dict == 'Layer1'):
                model.add(layer_type(neuron_num, input_dim=input_dim, kernel_initializer=kernel_initializer,
                                     activation=activation))

            else:
                if layer_name_asstr == 'Dense':
                    model.add(layer_type(neuron_num, kernel_initializer=kernel_initializer, activation=activation))
                if layer_name_asstr == 'Dropout':
                    model.add(layer_type(rate=rate))

        return model

    def fit(self, X, Y):
        # Need to rebuild and re-compile model at every fit instance so don't have information of weights from other fits
        self.model = self.build_model()
        self.model.compile(loss=self.loss, optimizer=self.optimizer, metrics=self.metrics)
        return self.model.fit(X, Y, epochs=self.epochs, batch_size=self.batch_size, verbose=self.verbose,
                              validation_split=self.validation_split, shuffle=self.shuffle)

    def predict(self, X):
        return self.model.predict(X)

    def summary(self):
        return self.model.summary()

# ref: https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.BaggingRegressor.html#sklearn.ensemble.BaggingRegressor
# NOTE: in order to use this, other models for the custom ensemble must be defined 
#       in the conf file with "_ensemble" somewhere in the name
class EnsembleRegressor():
    def __init__(self, n_estimators, num_samples, model_list, num_models):
        self.model_list = model_list # should be list of strings
        self.num_models = num_models # how many of each of the specified models should be included in the ensemble
        self.n_estimators = sum(self.num_models)
        self.num_samples = num_samples
        self.max_samples = num_samples
        self.bootstrapped_datasets = []
        self.bootstrapped_idxs = []
        self.all_preds = []
        self.path = ""
        self.model = self.build_models() # actually a list of models for use as the members in the ensemble

        self.fold = -1

        self.bootstrap = True

    def build_models(self):
        model = []

        for i, num_m in enumerate(self.num_models):
            for j in range(num_m):
                model.append(self.model_list[i])

        return model

    def setup(self, path):
        self.fold += 1
        self.bootstrapped_idxs = []
        self.bootstrapped_datasets = []
        self.path = path

    def fit(self, X, Y):
        X = X.values
        Y = Y.values

        idxs = np.arange(len(X))
        # fit each model in the ensemble
        for i in range(self.n_estimators):
            model = self.model[i]

            # do bootstrapping given the validation data
            bootstrap_idxs = random.choices(idxs, k=self.num_samples)
            bootstrap_X = X[bootstrap_idxs]
            bootstrap_Y = Y[bootstrap_idxs]
            if 1 == len(bootstrap_X.shape):
                bootstrap_X = np.expand_dims(np.asarray(bootstrap_X), -1)
            if 1 == len(bootstrap_Y.shape):
                bootstrap_Y = np.expand_dims(np.asarray(bootstrap_Y), -1)

            self.bootstrapped_idxs.append(bootstrap_idxs)
            self.bootstrapped_datasets.append(bootstrap_X)
            model.fit(bootstrap_X, bootstrap_Y)

    def predict(self, X, return_std=False):

        if isinstance(X, pd.DataFrame):
            X = X.values

        all_preds = []
        means = []

        for x_i in range(len(X)):
            preds = []
            for i in range(self.n_estimators):
                sample_X = X[x_i]
                if 1 == len(sample_X.shape):
                    sample_X = np.expand_dims(np.asarray(sample_X), 0)
                preds.append(self.model[i].predict(sample_X))
            all_preds.append(preds)
            means.append(np.mean(preds))

            # NOTE for ref (if manual jackknife implementation is necessary)
            # https://www.jpytr.com/post/random_forests_and_jackknife_variance/
            # https://github.com/scikit-learn-contrib/forest-confidence-interval/tree/master/forestci
            # http://contrib.scikit-learn.org/forest-confidence-interval/reference/forestci.html

        self.all_preds = all_preds

        return np.asarray(means)

    # check for failed fits, warn users, and re-calculate
    def stats_check_models(self, X, Y):
        if self.n_estimators > 10:
            maes = []
            for i in range(self.n_estimators):
                abs_errors = np.absolute(np.absolute(np.squeeze(np.asarray(self.all_preds)[:,i])) - Y)
                maes.append(sum(abs_errors) / len(abs_errors))

            alpha = 0.01
            bad_idxs = []
            for i in range(self.n_estimators):
                other_maes = np.delete(maes, [i])
                # ref: https://towardsdatascience.com/statistical-significance-hypothesis-testing-the-normal-curve-and-p-values-93274fa32687
                z_score = (maes[i] - np.mean(other_maes)) / np.std(other_maes)
                # ref: https://stackoverflow.com/questions/3496656/convert-z-score-z-value-standard-score-to-p-value-for-normal-distribution-in/3508321
                p_val = stats.norm.sf(abs(z_score))*2

                if p_val <= alpha:
                    # TODO ok to print these/how to print/log properly?
                    print("Estimator {} failed under statistical significance threshold {} (p_val {}), relevant dataset output to file with name format \'<fold>_<estimator idx>_bootstrapped_dataset.csv\'".format(i, alpha, p_val))
                    print("bad estimator mae: {}".format(maes[i]))
                    print("mean mae (for ref):")
                    print(np.mean(maes))
                    np.savetxt(self.path + "\\{}_{}_bootstrapped_dataset.csv".format(self.fold, i), self.bootstrapped_datasets[i], delimiter=",")
                    bad_idxs.append(i)

            if len(bad_idxs) == self.n_estimators:
                print("ALL models failed, wtf is your data")
                return
            #self.all_preds = np.delete(self.all_preds, bad_idxs, 1)

        y_preds = []
        for idx, x_i in enumerate(self.all_preds):
            y_preds.append(np.mean(x_i))

        return np.asarray(y_preds)

class ModelImport():
    """
    Class used to import pickled models from previous machine learning fits

    Args:

        model_path (str): string designating the path to load the saved .pkl model file

    Methods:

        fit: Does nothing, present for compatibility purposes

            Args:

                X: Nonetype

                y: Nonetype

                groups: Nonetype

        predict: Provides predicted model values based on X features

            Args:

                X: (numpy array), array of X features

            Returns:

                (numpy array), prediction array using imported model

    """
    def __init__(self, model_path):
        self.model_path = model_path
        self.model = joblib.load(self.model_path)

    def fit(self, X=None, y=None, groups=None):
        """ Only here for compatibility """
        return

    def predict(self, X):
        return self.model.predict(X)

# Optional to have xgboost working
try:
    custom_models = {
        'AlwaysFive': AlwaysFive,
        'RandomGuesser': RandomGuesser,
        'ModelImport': ModelImport,
        'XGBRegressor': xgb.XGBRegressor,
        'XGBClassifier': xgb.XGBClassifier,
        'KerasRegressor': KerasRegressor,
        'EnsembleRegressor': EnsembleRegressor
        #'DNNClassifier': keras_models.DNNClassifier
    }
except NameError:
    custom_models = {
        'AlwaysFive': AlwaysFive,
        'RandomGuesser': RandomGuesser,
        'ModelImport': ModelImport,
        'KerasRegressor': KerasRegressor,
        'EnsembleRegressor': EnsembleRegressor
        # 'DNNClassifier': keras_models.DNNClassifier
    }
name_to_constructor.update(custom_models)

def find_model(model_name):
    """
    Method used to check model names conform to scikit-learn model/estimator names

    Args:

        model_name: (str), the name of a model/estimator

    Returns:

        (str), the scikit-learn model name or raises InvalidModel error

    """
    try:
        return name_to_constructor[model_name]
    except KeyError:
        raise utils.InvalidModel(f"Model '{model_name}' does not exist in scikit-learn.")

def check_models_mixed(model_names):
    """
    Method used to check whether the user has mixed regression and classification tasks

    Args:

        model_names: (list), list containing names of models/estimators

    Returns:

        (bool), whether or not a classifier was found, or raises exception if both regression and classification models present.

    """

    found_classifier = found_regressor = False
    for name in model_names:
        if name in custom_models: continue
        class1 = find_model(name)
        if issubclass(class1, sklearn.base.ClassifierMixin):
            found_classifier = True
        elif issubclass(class1, sklearn.base.RegressorMixin):
            found_regressor = True
        else:
            raise Exception(f"Model '{name}' is neither a classifier nor a regressor")

    if found_classifier and found_regressor:
        raise Exception("Both classifiers and regressor models have been included")

    return found_classifier




"""
The data_cleaner module is used to clean missing or NaN values from pandas dataframes (e.g. removing NaN, imputation, etc.)
"""

import pandas as pd
import numpy as np
import logging
from sklearn.impute import SimpleImputer as Imputer

import os
from scipy.linalg import orth

log = logging.getLogger('mastml')

def flag_outliers(df, conf_not_input_features, savepath, n_stdevs=3):
    """
    Method that scans values in each X feature matrix column and flags values that are larger than 3 standard deviations
    from the average of that column value. The index and column values of potentially problematic points are listed and
    written to an output file.

    Args:
        df: (dataframe), pandas dataframe containing data

    Returns:
        None, just writes results to file

    """
    n_rows = df.shape[0]
    outlier_dict = dict()
    for col in df.columns:
        outlier_rows = list()
        outlier_vals = list()
        if col not in conf_not_input_features:
            avg = np.average(df[col])
            stdev = np.std(df[col])
            for row in range(n_rows):
                if df[col].iloc[row] > avg + n_stdevs*stdev:
                    outlier_rows.append(row)
                    outlier_vals.append(df[col].iloc[row])
                elif df[col].iloc[row] < avg - n_stdevs*stdev:
                    outlier_rows.append(row)
                    outlier_vals.append(df[col].iloc[row])
                else:
                    pass
        outlier_dict[col] = (outlier_rows, outlier_vals)
    pd.DataFrame().from_dict(data=outlier_dict,orient='index', columns=['Indices', 'Values']).to_excel(os.path.join(savepath,'data_potential_outliers.xlsx'))
    return

def remove(df, axis):
    """
    Method that removes a full column or row of data values if one column or row contains NaN or is blank

    Args:
        df: (dataframe), pandas dataframe containing data
        axis: (int), whether to remove rows (axis=0) or columns (axis=1)

    Returns:
        df: (dataframe): dataframe with NaN or missing values removed

    """
    df_nan = df[pd.isnull(df)]
    nan_indices = df_nan.index
    df = df.dropna(axis=axis, how='any')
    return df, nan_indices

def imputation(df, strategy, cols_to_leave_out=None):
    """
    Method that imputes values to the missing places based on the median, mean, etc. of the data in the column

    Args:
        df: (dataframe), pandas dataframe containing data
        strategy: (str), method of imputation, e.g. median, mean, etc.
        cols_to_leave_out: (list), list of column indices to not include in imputation

    Returns:
        df: (dataframe): dataframe with NaN or missing values resolved via imputation

    """
    col_names = df.columns.tolist()
    if cols_to_leave_out is None:
        df_imputed = pd.DataFrame(Imputer(missing_values='NaN', strategy=strategy, axis=0).fit_transform(df))
    else:
        df_include = df.drop(cols_to_leave_out, axis=1)
        df_hold_out = df.drop([c for c in df.columns if c not in cols_to_leave_out], axis=1)
        df_imputed = pd.DataFrame(Imputer(missing_values='NaN', strategy=strategy, axis=0).fit_transform(df_include), columns=df_include.columns)
    # Need to join the imputed dataframe with the columns containing strings that were held out
    if cols_to_leave_out is None:
        df = df_imputed
    else:
        df = pd.concat([df_hold_out, df_imputed], axis=1)
        col_names = df.columns.tolist()
    return df

def ppca(df, cols_to_leave_out=None):
    """
    Method that performs a recursive PCA routine to use PCA of known columns to fill in missing values in particular column

    Args:
        df: (dataframe), pandas dataframe containing data
        cols_to_leave_out: (list), list of column indices to not include in imputation

    Returns:
        df: (dataframe): dataframe with NaN or missing values resolved via imputation

    """
    col_names = df.columns.tolist()
    pca_magic = PPCA()
    if cols_to_leave_out is None:
        pca_magic.fit(np.array(df))
    else:
        pca_magic.fit(np.array(df.drop(cols_to_leave_out, axis=1)))
    # Need to un-standardize the pca-transformed data
    df_ppca = pd.DataFrame(pca_magic.data*pca_magic.stds+pca_magic.means)
    if cols_to_leave_out is None:
        df = df_ppca
    else:
        df = pd.concat([df_ppca, df[cols_to_leave_out]], axis=1)
    df.columns = col_names
    return df

def columns_with_strings(df):
    """
    Method that ascertains which columns in data contain string entries

    Args:
        df: (dataframe), pandas dataframe containing data

    Returns:
        str_columns: (list), list containing indices of columns containing strings

    """
    str_summary = pd.DataFrame(df.applymap(type).eq(str).any())
    str_columns = str_summary.index[str_summary[0] == True].tolist()
    return str_columns

class PPCA():
    """
    Class to perform probabilistic principal component analysis (PPCA) to fill in missing data.

    This PPCA routine was taken directly from https://github.com/allentran/pca-magic. Due to import errors, for ease of use
    we have elected to copy the module here. This github repo was last accessed on 8/27/18. The code comprising the PPCA
    class below was not developed by and is not owned by the University of Wisconsin-Madison MAST-ML development team.

    """
    def __init__(self):

        self.raw = None
        self.data = None
        self.C = None
        self.means = None
        self.stds = None
        self.eig_vals = None

    def _standardize(self, X):

        if self.means is None or self.stds is None:
            raise RuntimeError("Fit model first")

        return (X - self.means) / self.stds

    def fit(self, data, d=None, tol=1e-4, min_obs=10, verbose=False):

        self.raw = data
        self.raw[np.isinf(self.raw)] = np.max(self.raw[np.isfinite(self.raw)])

        valid_series = np.sum(~np.isnan(self.raw), axis=0) >= min_obs

        data = self.raw[:, valid_series].copy()
        N = data.shape[0]
        D = data.shape[1]

        self.means = np.nanmean(data, axis=0)
        self.stds = np.nanstd(data, axis=0)

        data = self._standardize(data)
        observed = ~np.isnan(data)
        missing = np.sum(~observed)
        data[~observed] = 0

        # initial

        if d is None:
            d = data.shape[1]

        if self.C is None:
            C = np.random.randn(D, d)
        else:
            C = self.C
        CC = np.dot(C.T, C)
        X = np.dot(np.dot(data, C), np.linalg.inv(CC))
        recon = np.dot(X, C.T)
        recon[~observed] = 0
        ss = np.sum((recon - data) ** 2) / (N * D - missing)

        v0 = np.inf
        counter = 0

        while True:

            Sx = np.linalg.inv(np.eye(d) + CC / ss)

            # e-step
            ss0 = ss
            if missing > 0:
                proj = np.dot(X, C.T)
                data[~observed] = proj[~observed]
            X = np.dot(np.dot(data, C), Sx) / ss

            # m-step
            XX = np.dot(X.T, X)
            C = np.dot(np.dot(data.T, X), np.linalg.pinv(XX + N * Sx))
            CC = np.dot(C.T, C)
            recon = np.dot(X, C.T)
            recon[~observed] = 0
            ss = (np.sum((recon - data) ** 2) + N * np.sum(CC * Sx) + missing * ss0) / (N * D)

            # calc diff for convergence
            det = np.log(np.linalg.det(Sx))
            if np.isinf(det):
                det = abs(np.linalg.slogdet(Sx)[1])
            v1 = N * (D * np.log(ss) + np.trace(Sx) - det) \
                 + np.trace(XX) - missing * np.log(ss0)
            diff = abs(v1 / v0 - 1)
            if verbose:
                print(diff)
            if (diff < tol) and (counter > 5):
                break

            counter += 1
            v0 = v1

        C = orth(C)
        vals, vecs = np.linalg.eig(np.cov(np.dot(data, C).T))
        order = np.flipud(np.argsort(vals))
        vecs = vecs[:, order]
        vals = vals[order]

        C = np.dot(C, vecs)

        # attach objects to class
        self.C = C
        self.data = data
        self.eig_vals = vals
        self._calc_var()

    def transform(self, data=None):

        if self.C is None:
            raise RuntimeError('Fit the data model first.')
        if data is None:
            return np.dot(self.data, self.C)
        return np.dot(data, self.C)

    def _calc_var(self):

        if self.data is None:
            raise RuntimeError('Fit the data model first.')

        data = self.data.T

        # variance calc
        var = np.nanvar(data, axis=1)
        total_var = var.sum()
        self.var_exp = self.eig_vals.cumsum() / total_var

    def save(self, fpath):

        np.save(fpath, self.C)

    def load(self, fpath):

        assert os.path.isfile(fpath)

        self.C = np.load(fpath)



"""
The clusterers module is used for instantiating cluster algorithm objects from scikit-learn.
More information is available at http://scikit-learn.org/stable/modules/classes.html#module-sklearn.cluster
"""

import sklearn.cluster as sc

name_to_constructor = {
    'AffinityPropagation': sc.AffinityPropagation, # Perform Affinity Propagation Clustering of data.
    'AgglomerativeClustering': sc.AgglomerativeClustering, # Agglomerative Clustering
    'Birch': sc.Birch, # Implements the Birch clustering algorithm.
    'DBSCAN': sc.DBSCAN, # Perform DBSCAN clustering from vector array or distance matrix.
    'FeatureAgglomeration': sc.FeatureAgglomeration, # Agglomerate features.
    'KMeans': sc.KMeans, # K-Means clustering
    'MiniBatchKMeans': sc.MiniBatchKMeans, # Mini-Batch K-Means clustering
    'MeanShift': sc.MeanShift, # Mean shift clustering using a flat kernel.
    'SpectralClustering': sc.SpectralClustering, # Apply clustering to a projection to the normalized laplacian.
    }



"""
The conf_parser module is used for handling, parsing, and checking MAST-ML input configuration files
"""

from sklearn.metrics import make_scorer
from configobj import ConfigObj
import logging

from mastml import metrics, utils
from mastml.legos.model_finder import check_models_mixed
from mastml.legos import feature_selectors, model_finder

log = logging.getLogger('mastml')

def parse_conf_file(filepath, from_dict=False):
    """
    Method that accepts the filepath of an input configuration file and returns its parsed dictionary

    Args:
        filepath: (str), path to config file, or a dict of config values directly

    Returns:
        conf: (dict): dictionary parsed from config file

    """

    if from_dict == False:
        conf = ConfigObj(filepath)
    else:
        conf = filepath # The filepath in this case is an actual dictionary of values used as the config fiel

    main_sections = ['GeneralSetup', 'DataSplits', 'Models', 'LearningCurve', 'DataCleaning', 'HyperOpt', 'ModelHosting']
    feature_sections = ['FeatureGeneration', 'Clustering',
                        'FeatureNormalization', 'FeatureSelection']
    feature_section_dicts = [conf[name] for name in feature_sections if name in conf]

    def set_required_sections_to_empty():
        for name in main_sections:
            if name not in conf:
                conf[name] = dict()
    set_required_sections_to_empty()

    def check_unknown_sections():
        all_sections = main_sections + feature_sections + ['MiscSettings']
        for section_name in conf:
            if section_name not in all_sections:
                raise Exception(f'[{section_name}] is not a valid section!'
                                f' Valid sections: {all_sections}')
    check_unknown_sections()

    def verify_subsection_only_sections():
        for dictionary in [conf, conf['DataSplits'], conf['Models']] + feature_section_dicts:
            for name, value in dictionary.items():
                if not isinstance(value, dict):
                    raise utils.InvalidConfSubSection(
                            f"Parameter in subsection-only section: {name}={value}")
    verify_subsection_only_sections()

    def parameter_dict_type_check_and_cast():
        parameter_dicts = list()
        parameter_dicts.extend(conf['Models'].values())
        parameter_dicts.extend(conf['DataSplits'].values())
        #parameter_dicts = conf['Models'].values() + conf['DataSplits'].values()
        for feature_section in feature_section_dicts:
            parameter_dicts.extend(feature_section.values())

        for parameter_dict in parameter_dicts:
            for name, value in parameter_dict.items():
                # Does this parameter-only section include a subsection?
                if isinstance(value, dict):
                    continue
                    #raise utils.InvalidConfSubSection(
                    #        f"Subsection in parameter-only section: {key}")
                # cast the strings to their respective types
                parameter_dict[name] = fix_types(value)
    parameter_dict_type_check_and_cast()

    # Ensure all models are either classifiers or regressors: (raises error if mixed)
    is_classification = conf['is_classification'] = check_models_mixed(
            key.split('_')[0] for key in conf['Models'])

    # Add the empty splitter if no splitters are specified:
    if conf['DataSplits'] == dict():
        conf['DataSplits']['NoSplit'] = dict()

    def set_unspecified_sections_to_empty_dict():
        for name in feature_sections:
            if name not in conf or conf[name] == dict():
                if name == 'Clustering':
                    conf[name] = dict()
                else:
                    conf[name] = {'DoNothing': dict()}
    set_unspecified_sections_to_empty_dict()

    GS = conf['GeneralSetup']

    def check_general_setup_settings_are_valid():
        #all_settings =  ['input_features', 'target_feature', 'metrics',
        #                 'randomizer', 'validation_columns', 'not_input_features', 'grouping_feature']
        all_settings =  ['input_features', 'input_target', 'metrics',
                         'randomizer', 'input_testdata', 'input_other', 'input_grouping']
        for name in GS:
            if name not in all_settings:
                raise utils.InvalidConfParameters(
                        f"[GeneralSetup] contains unknown setting {name}.\n"
                        f"Valid GeneralSetup options are: {all_settings}")
    check_general_setup_settings_are_valid()

    if 'input_grouping' not in GS:
        GS['input_grouping'] = None

    # Find grouping features and 'not_input_features' to blacklist out of X (see data loader)
    def collect_grouping_features():
        for section in conf:
            if not isinstance(conf[section], dict):
                continue
            for subsection in conf[section]:
                SS = conf[section][subsection]
                if not isinstance(SS, dict):
                    continue
                if 'input_grouping' in SS.keys():
                    logging.debug('found input_grouping feature: ' + SS['input_grouping'])
                    yield SS['input_grouping']
    # Issue here where if clusters are automatically generated, new column is made but isn't in intitial df, even though
    # listed as grouping_feature. Here, just have to remember to put grouping_feature names in not_input_features
    #feature_blacklist = list(collect_grouping_features())
    feature_blacklist = list()
    # default not_input_features to a list
    if 'input_other' not in GS:
        GS['input_other'] = list()
    else:
        if type(GS['input_other']) is str:
            new_list = list()
            new_list.append(GS['input_other'])
            GS['input_other'] = new_list
        elif type(GS['input_other']) is list:
            pass

    # and add the discovered ones to the list
    GS['input_other'] += feature_blacklist
    #GS['not_input_features'] = [f for f in feature_blacklist if f not in GS['not_input_features']]

    def set_randomizer_setting():
        if 'randomizer' in GS:
            GS['randomizer'] = mybool(GS['randomizer'])
        else:
            GS['randomizer'] = False
    set_randomizer_setting()


    def set_default_features():
        for name in ['input_features', 'input_target']:
            if (name not in GS) or (GS[name] == 'Auto'):
                GS[name] = None
    set_default_features()

    def set_default_metrics():
        if 'metrics' not in GS or GS['metrics'] == 'Auto':
            if is_classification:
                GS['metrics'] = ['accuracy', 'precision_binary', 'recall_binary', 'f1_binary']
            else:
                GS['metrics'] = ['R2', 'root_mean_squared_error',
                                 'mean_absolute_error', 'rmse_over_stdev']
    set_default_metrics()

    # Turn names of metrics into actual metrics:
    # If only one metric, map string to list for parsing
    if type(GS['metrics']) is str:
        el = GS['metrics']
        GS['metrics'] = list()
        GS['metrics'].append(el)

    GS['metrics'] = metrics.check_and_fetch_names(GS['metrics'], is_classification)

    def change_score_func_strings_into_actual_score_funcs():
        for selector_name, args_dict in conf['FeatureSelection'].items():
            class_name = selector_name.split('_')[0]
            if class_name not in feature_selectors.score_func_selectors: continue
            name_to_func = (metrics.classification_score_funcs
                            if is_classification
                            else metrics.regression_score_funcs)
            if 'score_func' in args_dict:
                try:
                    args_dict['score_func'] = name_to_func[args_dict['score_func']]
                except KeyError:
                    task = 'classification' if is_classification else 'regression'
                    raise utils.InvalidValue(
                            f"Score function '{args_dict['score_func']}' not valid for {task}"
                            f" tasks (inside feature selector {selector_name}). Valid score"
                            f" functions: {list(name_to_func.keys())}")
            else: # default to f_classif or f_regression
                args_dict['score_func'] = \
                        name_to_func['f_classif' if is_classification else 'f_regression']
    change_score_func_strings_into_actual_score_funcs()

    def make_long_name_short_name_pairs():
        dictionaries = ([conf['DataSplits'], conf['Models'], conf['HyperOpt']]
                        + [conf[name] for name in feature_sections])
        for dictionary in dictionaries:
            for name, settings in dictionary.items():
                #dictionary[name] = (name.split('_')[0], settings)
                dictionary[name] = [name.split('_')[0], settings]
    make_long_name_short_name_pairs()

    def check_and_boolify_plot_settings():
        default_false = ['plot_each_feature_vs_target', 'rf_error_method', 'rf_error_percentile',
                         'normalize_target_feature']
        default_true = ['plot_target_histogram', 'plot_train_test_plots', 'plot_predicted_vs_true', 'plot_error_plots',
                         'plot_predicted_vs_true_average', 'plot_best_worst_per_point']
        all_settings = default_false + default_true
        if 'MiscSettings' not in conf:
            conf['MiscSettings'] = dict()
        MS = conf['MiscSettings']
        for name, value in MS.items():
            if name not in all_settings:
                raise utils.InvalidConfParameters(f"[MiscSettings] parameter '{name}' is unknown")
            try:
                MS[name] = mybool(value)
            except ValueError:
                pass
            #    raise utils.InvalidConfParameters(
            #        f"[PlotSettings] parameter '{name}' must be a boolean")
        for name in default_false:
            if name not in MS:
                MS[name] = False
        for name in default_true:
            if name not in MS:
                MS[name] = True
    check_and_boolify_plot_settings()

    # TODO: remove?
    def check_learning_curve_settings():
        if 'learning_curve_model' not in GS:
            raise utils.InvalidConfParameters("You enabled data_learning_curve plots but you did"
                                              "not specify learning_curve_model in [GeneralSetup]")
        if 'learning_curve_score' not in GS:
            raise utils.InvalidConfParameters("You enabled data_learning_curve plots but you did"
                                              "not specify learning_curve_score in [GeneralSetup]")

    # Need to make scoring function for learning curve string to scorer object
    if conf['LearningCurve']:
        score_name = conf['LearningCurve']['scoring']
        d = metrics.check_and_fetch_names([score_name], is_classification)
        greater_is_better, score_func = d[score_name]
        conf['LearningCurve']['scoring'] = make_scorer(score_func, greater_is_better=True)

    return conf

def fix_types(maybe_list):
    """
    Method that returns true datatype of values passed as string or list of strings, parsed from configuration file

    Args:
        maybe_list: (list, str), a list of strings or just a string whose datatype should be e.g. int or list of float

    Returns:
        maybe_list: (list, bool, int, float): a list of items or other data type converted from string to correct data type

    """
    if isinstance(maybe_list, list):
        return [fix_types(item) for item in maybe_list]

    try: return mybool(maybe_list)
    except ValueError: pass

    try: return int(maybe_list)
    except ValueError: pass

    try: return float(maybe_list)
    except ValueError: pass

    return str(maybe_list)

def mybool(string):
    """
    Method that converts a string equal to 'True' or 'False' into type bool

    Args:
        string: (str), a string as 'True' or 'False'

    Returns:
        bool: (bool): bool as True or False

    """
    if string.lower() == 'true':
        return True
    if string.lower() == 'false':
        return False
    raise ValueError



"""
This module contains a class used to randomize the input y data, in order to create a "null model" for testing how
rigorous other machine learning model predictions are.
"""

class Randomizer():
    """
    Class which randomizes X-y pairings by shuffling the y values

    Args:

        None

    Methods:

        fit: just passes through; present to maintain scikit-learn structure

            Args:

                None

        transform: randomizes the values of a dataframe

            Args:

                df: (dataframe), a dataframe with data to be randomized

            Returns:

                (dataframe), a dataframe with randomized data

    """

    def __init__(self):
        pass

    def fit(self):
        return self

    def transform(self, df):
        return df.sample(frac=1).reset_index(drop=True)